#!/bin/sh
#	This is a shell archive.
#	Run the file through sh to extract its contents.
# shar:	Shell Archiver
#	Run the following text with /bin/sh to create:
#	scanner.s
#	ll-parsers.s
#	grammars.s
#	Makefile
#	README
#	NEWS
# This archive created: Tue Nov  7 11:24:44 1989
echo shar: extracting scanner.s '(8403 characters)'
sed 's/^X//' << \SHAR_EOF > scanner.s
X;;; Scanner Generator
X
X;;; Source:  Book, Sec 3.1, modified for Scheme84.
X;;; Modified for Chez Scheme Fri Sep 30 11:00:18 1988
X
X;;; ******************************************************************
X
X;;; data structure definitions
X
X;;; for automata    
X
X;;; automaton ::= (state*)
X;;; state ::= (state-name action transition*)
X;;; state-name ::= symbol
X;;; action ::= expression | bogus
X;;; transition ::= (input-spec state-name)
X;;; input-spec ::= char | (char char)
X
X;;; the automaton has two special states named initial, and end-marker.
X;;; initial is the initial state.
X;;; end-marker is a state with no transitions; its action is used to
X;;; produce and end-marker item in the resulting list of values.
X
X(define state->name car)
X(define state->action cadr)
X(define state->transitions caddr)
X
X(define transition->input-spec car)
X(define transition->next-state cadr)
X
X;;; items  -- used internally to pass items and remainders around
X
X;;; item ::= (*bogus unused-chars)
X;;; item ::= (*ordinary unused-chars value)
X
X(define make-bogus-item (lambda (chars) (list '*bogus chars)))
X(define make-real-item
X  (lambda (value chars) (list '*ordinary  chars value)))
X
X(define item->unused cadr)
X(define item->value caddr)
X
X;;; lexical-item -- used externally to pass items to parser
X
X;;; lexical-item ::= (class data)
X;;; value is () for a token
X
X(define lexical-item->class car)
X(define lexical-item->data cadr)
X(define make-lexical-item (lambda (class data) (list class data)))
X
X;;; *************************************************************
X
X;;; Top-level macro-- define-scanner
X
X;;; the actions are evaluated at scanner-construction time, everything
X;;; else is not.
X
X(extend-syntax (define-scanner)
X  [(define-scanner name
X     (state-name action transition ...) ...)
X   (define name
X     (let ([automaton
X	     (list (list 'state-name
X			 (lambda (chars) (action chars)) ; action is
X							 ; dynamically bound.
X			 '(transition ...)) ...)])
X       (set! automaton-1 automaton)
X       (lambda (string) (scan-string automaton string))))])
X     
X;;; ****************************************************************  
X
X;;; Top-level Function--  scan-string 
X
X;;; scan-string:  takes an automaton, a string, and an end-marker, and
X;;; returns a list of items (determined by the actions in the
X;;; automaton), terminated by the end-marker item.
X
X(define scan-string
X  (lambda (automaton input-string)
X    (let loop ([index 0])
X      (if (>= index (string-length input-string))
X	(list (end-marker-action automaton))
X	(let ([item (next-item-raw automaton input-string index)])
X	  ;;(printf "scan-string: item = ~s~%" item)
X	  (case (car item)
X	    [(*bogus) (loop (item->unused item))]
X	    [(*ordinary)
X	     (cons (item->value item)
X		   (loop (item->unused item)))]))))))
X
X;;;**********************************************************************
X
X;;; Simulate Automaton
X
X;;; Given a list, a predicate, and a failure value, return the first
X;;; element in the list that satisfies the predicate.  If no such
X;;; element exists, return #f
X
X(define linear-search
X  (lambda (l predicate)
X    (cond
X      [(null? l) #f]
X      [(predicate (car l)) (car l)]
X      [else (linear-search (cdr l) predicate)])))
X
X(define find-state
X  (lambda (automaton state-name)
X    (linear-search
X      automaton
X      (lambda (s) (eq? (state->name s) state-name)))))
X
X;;; Given a state and an input-char, find the next state-name if any.
X;;; Otherwise return #f. 
X
X(define find-transition
X  (lambda (state input-char)
X    '(printf "find-transition: state = ~s input-char = ~s~%"
X       state input-char)
X    (linear-search
X      (state->transitions state)
X      (lambda (transition)
X	'(printf "transition: ~s input-char: ~s~%" transition input-char)
X	(let ([input-spec (transition->input-spec transition)])
X	  '(printf "find-transition: input-spec = ~s input-char = ~s~%"
X		 input-spec input-char)
X	       (cond
X		 ((char? input-spec)	
X		  ;; it's a single character:
X		  (eqv? input-char input-spec))
X		 ;; no, it's a subrange
X		 (else
X		   (let ((ans (char<=?
X				(car input-spec)
X				input-char
X				(cadr input-spec))))
X		     ans))))))))
X
X
X;;; simulate automaton and collect characters.
X
X(define next-item-raw
X  (lambda (automaton input-string index)
X    (let main-loop
X      ([state-name 'initial][char-buffer '()][index index])
X      (let ([state (find-state automaton state-name)])
X	'(printf "next-item-raw: state-name = ~s state = ~s~%"
X	  state-name state)
X	(cond
X	  [(null? state)
X	   ;; bad state-name
X	   (error 'next-item-raw "can't find state named ~s" state-name)]
X	  [(>= index (string-length input-string))
X	   ;; this is the end of the string, so it must be the of
X	   ;; the item.
X	   '(printf "next-item-raw: found end of string in state ~s~%"
X	      state-name)
X	   (make-item 
X	     (state->action state)
X	     (reverse char-buffer)
X	     index)]
X	  [else
X	    (let* ([char (string-ref input-string index)]
X		   [transition
X		     (find-transition state char)])
X	      '(printf "state-name: ~s char: ~s~%" state-name char)
X	      '(printf "transition: ~s~%" transition)
X	      (cond
X		;; if we've found the next transition, accumulate the
X		;; character in the char-buffer, and go on to the next
X		;; character:
X		[(not (eq? transition #f))
X		 (main-loop
X		   (transition->next-state transition)
X		   (cons char char-buffer)
X		   (+ index 1))]
X		;; no transition applies, so we must be looking at
X		;; the first character of the next item
X		[else
X		  (make-item
X		    (state->action state)
X		    (reverse char-buffer)
X		    index)]))])))))
X
X(define make-item
X  (lambda (action buffer unused)
X    (let ((value (action buffer)))
X      (if (eq? value 'bogus)
X	(make-bogus-item unused)
X	(make-real-item value unused)))))
X
X
X(define end-marker-action
X  (lambda (automaton)
X    (let ([state (find-state automaton 'end-marker)])
X      (if state
X	((state->action state) '())
X	(error 'end-marker-action
X	    "can't find state named end-marker")))))
X
X;;; **********************************************************************
X
X;;; Example:
X
X(define cook-symbol
X  (lambda (chars)
X    (let ([sym (char-list->symbol chars)])
X      (if (memq sym **keywords-list**)
X             (make-lexical-item sym '())
X             (make-lexical-item 'symbol sym))))) ; this will change
X						  ; to symbol eventually
X
X(define **keywords-list** '(let in if then else set!))
X
X(define cook-bogus
X  (lambda (chars) 'bogus))
X
X(define cook-number
X  (lambda (chars)
X    (make-lexical-item 'int (digit-list->number chars))))
X
X(define lex-error
X  (lambda (chars)
X    (error 'lex-error "unknown lexical item ~s" chars)))
X
X(define cook-token
X  (lambda (token)
X    (lambda (chars)
X      (make-lexical-item token '()))))
X
X(define char-list->symbol 
X  (lambda (char-list) (string->symbol (list->string char-list))))
X
X;;; digit-list->number converts a list of decimal digits to a number.
X;;; It uses the fact that the ASCII characters for 0..9 are
X;;; consecutive, starting with 0.
X
X(define digit-list->number
X  (let ([code-for-zero (char->integer #\0)])
X    (lambda (char-list)
X      (letrec
X       ([loop (lambda (accum char-list)
X                (if (null? char-list)
X                    accum
X                    (loop
X                     (+ (* 10 accum)
X                        (- (char->integer (car char-list))
X                           code-for-zero))
X                     (cdr char-list))))])
X       (loop 0 char-list)))))
X
X
X
X;;; *******************************************************************
X
X;;; Reasonable lexical spec:
X
X(define-scanner scanner-1
X  (initial lex-error
X    (#\space whitespace)
X    (#\tab  whitespace)
X    (#\newline whitespace)
X    ((#\a #\z) symbol)
X    ((#\A #\Z) symbol)
X    (#\, comma)
X    ((#\* #\/) symbol)		; except "," 
X    (#\= symbol)
X    (#\( lparen)
X    (#\) rparen)
X    (#\% comment)
X    (#\; semicolon)
X    ((#\0 #\9) number))
X  (symbol cook-symbol
X    ((#\a #\z) symbol)
X    ((#\A #\Z) symbol)
X    ((#\0 #\9) symbol)
X    (#\* symbol)
X    (#\+ symbol)
X    (#\- symbol)
X    (#\/ symbol)
X    (#\! symbol)
X    (#\= symbol))
X  (comment cook-bogus
X    ((#\space #\~) comment))
X  (whitespace cook-bogus
X    (#\space whitespace)
X    (#\tab whitespace)
X    (#\newline whitespace))
X  (number cook-number
X    ((#\0 #\9)number))
X  (comma (cook-token 'comma))
X  (lparen (cook-token 'lparen))
X  (rparen (cook-token 'rparen))
X  (semicolon (cook-token 'semicolon))
X  (end-marker (cook-token 'end-marker)))
X
X
X
X
X
SHAR_EOF
if test 8403 -ne "`wc -c scanner.s`"
then
echo shar: error transmitting scanner.s '(should have been 8403 characters)'
fi
echo shar: extracting ll-parsers.s '(14401 characters)'
sed 's/^X//' << \SHAR_EOF > ll-parsers.s
X;;; Generate and interpret LL parsing tables
X
X;;; Assumes scanner is loaded
X
X;;; ***************************************************************
X
X;;; Top level: (sample)
X
X'(define parse
X  (let ([table (build-LL-table 'command grammar-1)])
X    (lambda (string)
X      (parse-whole table 'command
X	(scan-string automaton-1 cooking-function-1 string)))))
X
X
X'(define parse-whole
X  (lambda (table non-terminal item-list)
X    (let ([result (parse-nt table non-terminal item-list)])
X      (if (eq? (lexical-item->class
X		 (car (parser-result->unused result)))
X	       'end-marker)
X	  (parser-result->tree result)
X	  (error 'parse-top-level
X	    "symbols left over: ~s"
X	    (car (parser-result->unused result)))))))
X
X;;; *************************************************************
X
X;;; Extraction operators for grammars:
X
X;;; production ::= (lhs rhs action)
X
X(define production->lhs  car)	
X(define production->rhs  cadr)
X(define production->action caddr)
X
X;;; rhs ::= (rhs-entry ...)
X;;; rhs-entry ::= (token item-class) or (data-item item-class) or
X;;; (non-terminal non-terminal)
X
X(define rhs-entry->item-class    cadr)
X(define rhs-entry->tag  car)
X
X;;; ****************************************************************
X
X;;; Creating LL tables
X
X;;; Our general approach is generate first and follow sets by treating
X;;; the grammar as a set of axioms and inference rules:
X
X;;; Assertions:
X;;; assertion ::= (first non-term term)  means term is in first(non-term).
X;;; assertion ::= (follow non-term term) --similar
X
X;;; rule ::= ((first B) (follow A)) --means (first B x) => (follow A x)
X
X(define build-LL-table
X  (lambda (start-symbol grammar)
X    (let* ([rules&axioms (grammar->rules start-symbol grammar)]
X	   [rules (rules&axioms->rules rules&axioms)]
X	   [axioms (rules&axioms->axioms rules&axioms)]
X	   [assertions (deductive-closure apply-rule rules axioms)])
X      (assemble-LL-table assertions grammar))))
X
X(define grammar->rules
X  (lambda (start-symbol grammar)
X    (cons
X      `(follow ,start-symbol end-marker)
X      (apply append 
X	     (map production->rules grammar)))))
X
X;;; Here's where we generate the rules:
X
X(define production->rules
X  (lambda (production)
X    (let ([a (production->lhs production)]
X	  [rhs (production->rhs production)]
X	  [complain (lambda (rhs)
X		      (error 'production->rules
X			"can't treat entry ~s"
X			(car rhs)))])
X      (letrec
X	([loop (lambda (rhs)
X		 ;; assume rhs is non-empty
X		 '(printf "inner loop: rhs = ~s~%" rhs)
X		 (if
X		   (null? (cdr rhs))
X		   ;; normal termination for this loop
X		   (record-case (car rhs)
X		     [non-terminal (b)
X		       (list `((follow ,b) (follow ,a)))]
X		     [data-item (i) '()]
X		     [token (i) '()]
X		     [else (complain rhs)])
X		   (record-case (car rhs)
X		     [non-terminal (b)
X		       (record-case (cadr rhs)
X			 [non-terminal (c)
X			   (cons `((first ,c) (follow ,b))
X				 (loop (cdr rhs)))]
X			 [data-item (x)
X			   (cons `(follow ,b ,x) (loop (cdr rhs)))]
X			 [token (x)
X			   (cons `(follow ,b ,x) (loop (cdr rhs)))]
X			 [else (complain (cdr rhs))])]
X		     [data-item (i) (loop (cdr rhs))]
X		     [token (i) (loop (cdr rhs))]
X		     [else (complain rhs)])))])
X	(let ([first-rule
X		(if
X		  (null? rhs)
X		  `((follow ,a) (first ,a))
X		  (record-case (car rhs)
X		    [non-terminal (b) `((first ,b) (first ,a))]
X		    [data-item (x) `(first ,a ,x)]
X		    [token (x) `(first ,a ,x)]
X		    [else  (complain rhs)]))])
X	  (if (null? rhs)
X	    (list first-rule)
X	    (cons first-rule (loop rhs))))))))
X	    
X;;; Exercise:  extend this to handle while-lookahead-not and
X;;; parse-separated-list 
X		 
X(define rules&axioms->rules
X  (lambda (l)
X    (cond
X      [(null? l) '()]
X      [(symbol? (caar l)) (rules&axioms->rules (cdr l))]
X      [else (cons (car l) (rules&axioms->rules (cdr l)))])))
X
X(define rules&axioms->axioms
X  (lambda (l)
X    (cond
X      [(null? l) '()]
X      [(symbol? (caar l))
X       (cons (car l) (rules&axioms->axioms (cdr l)))]
X      [else
X	(rules&axioms->axioms (cdr l))])))
X
X(define apply-rule
X  (lambda (rule assertion succeed fail)
X    ;; rule ::= ((tag1 nt1) (tag2 nt2))
X    ;; assertion  ::= (tag nt x)
X    (if (and
X	  (eq? (caar rule) (car assertion))
X	  (eq? (cadar rule) (cadr assertion)))
X      (succeed
X	`(,(caadr rule) ,(cadr (cadr rule)) ,(caddr assertion)))
X      (fail))))
X
X(define deductive-closure
X  (lambda (apply-rule rules axioms)
X    (let outer-loop ([old '()][new axioms])
X      (if (null? new)
X	old
X	(let assertion-loop ([newer '()][assertions new])
X	  (if (null? assertions)
X	    (outer-loop (append new old) newer)
X	    (let rule-loop ([newer newer][rules rules])
X	      (if (null? rules)
X		(assertion-loop newer (cdr assertions))
X		(apply-rule (car rules) (car assertions)
X		  (lambda (conclusion)
X		    (if (already-deduced? conclusion
X			  newer new old)
X		      (rule-loop newer (cdr rules))
X		      (rule-loop (cons conclusion newer)
X			(cdr rules))))
X		  (lambda () (rule-loop newer (cdr rules))))))))))))
X
X;;; this should be replaced by some kind of hash table
X
X(define already-deduced?
X  (lambda (conclusion newer new old)
X    (or (member conclusion newer)
X	(member conclusion new)
X	(member conclusion old))))
X
X(define filter
X  (lambda (p l)
X    (cond
X      [(null? l) '()]
X      [(p (car l)) (cons (car l)
X			 (filter p (cdr l)))]
X      [else (filter p (cdr l))])))
X
X;;; Now we start constructing the LL tables
X;;; LL tables look just like productions, except we have
X;;;
X;;; lhs ::= (nt list-of-terminals)
X;;;
X;;; The list of terminals tells find-production which terminals this
X;;; production is applicable to.
X
X(define assemble-LL-table
X  (lambda (assertions grammar)
X    (map
X      (lambda (production)
X	(assemble-LL-production assertions production))
X      grammar)))
X
X(define assemble-LL-production
X  (lambda (assertions production)
X    (let ([lhs (production->lhs production)]
X	  [rhs (production->rhs production)]
X	  [action (production->action production)])
X      `((,lhs ,(extract-leading-terminals lhs rhs assertions))
X	,rhs
X	,action))))
X
X(define filter-assertions
X  (lambda (tag non-terminal assertions)
X    (map caddr
X	 (filter
X	   (lambda (assertion)
X	     (and (eq? (car assertion) tag)
X		  (eq? (cadr assertion) non-terminal)))
X	   assertions))))
X
X(define extract-leading-terminals
X  (lambda (lhs rhs assertions)
X    (if (null? rhs)
X      (filter-assertions 'follow lhs assertions)
X      (record-case (car rhs)
X	[non-terminal (nt)
X	  (filter-assertions 'first nt assertions)]
X	[token (class) (list class)]
X	[data-item (class) (list class)]
X	[else (error 'extract-leading-terminals
X		"unsupported rhs entry ~s" (car rhs))]))))
X
X  
X;;; ****************************************************************
X
X;;; Interpretation of LL tables
X
X;;; Parser-result:
X
X(define make-parser-result
X  (lambda (tree unused)
X    (list tree unused)))
X
X(define parser-result->tree    car)
X(define parser-result->unused  cadr)
X
X;;; For making records:
X
X(define tag&fields->record     cons)
X
X;;; *************************************************************
X
X;;; Procedures for interpreting parse tables:
X
X;;; The engine for interpeting parse tables is unchanged from the one
X;;; in the book (as revised), except that find-production has changed
X;;; to use the LL left sides.
X
X
X(define parse-nt
X  (lambda (grammar nt item-list)
X    ;; find production relevant to this nt and (car item-list), and
X    ;; call parse-production to process it.
X    (let ([production
X	    (find-production grammar nt
X	      (lexical-item->class (car item-list)))])
X      (if production
X	  (parse-production grammar production item-list)
X	  (error 'parse-nt
X	    "Nonterminal ~s can't start with ~s:~%~s"
X	    nt (lexical-item->class (car item-list))
X	    (car item-list)
X	    )))))
X
X(define find-production
X  (lambda (grammar non-terminal item-class)
X    (linear-search grammar
X      (lambda (production)
X	(let ([lhs (production->lhs production)])
X	  '(printf "find-production: lhs = ~s~%" lhs)
X	  (and (eq? (car lhs) non-terminal)
X	       (memq item-class (cadr lhs))))))))
X
X
X(define parse-production
X  (lambda (grammar production item-list)
X    ;; production ::= (lhs rhs action).  Call parse-rhs to process rhs.
X    ;; This returns a parser-result containing a list of subtrees.
X    ;; Last, create a parser-result containing the result of applying
X    ;; the action to the list of subtrees.
X    (let ([result (parse-rhs grammar
X		    (production->rhs production)
X		    item-list)]) 
X      (make-parser-result
X	(apply
X	  (production->action production)
X	  (parser-result->tree result))
X	(parser-result->unused result)))))
X
X(define parse-rhs
X  (lambda (grammar rhs item-list)
X    '(printf "parse-rhs: rhs = ~s ; (car item-list) = ~s~%"
X      rhs (car item-list))
X    (if (null? rhs)
X	(make-parser-result '() item-list)
X	(record-case (car rhs)
X	  ;; first entry is (token item-class).  Check to see if first
X	  ;; item matches the item-class.  If so, parse (cdr
X	  ;; item-list) against (cdr rhs).
X	  [token (item-class)
X	    (if (eq? (lexical-item->class (car item-list))
X		     item-class)
X	      (parse-rhs grammar (cdr rhs) (cdr item-list))
X	      (error 'parse-rhs "Looking for: ~s found: ~s"
X		item-class
X		(lexical-item->class (car item-list))))]
X	  ;; first entry is (data-item item-class).  Call
X	  ;; parse-data-item to check the first item, and pass the
X	  ;; result to finish-rhs, which will finish parsing the rhs.
X	  [data-item (item-class)
X	    (finish-rhs
X	      (parse-data-item item-class item-list)
X	      grammar (cdr rhs))]
X	  ;; (non-terminal nt) -- similar
X	  [non-terminal (nt)
X	    (finish-rhs
X	      (parse-nt grammar nt item-list)
X	      grammar (cdr rhs))]
X	  [while-lookahead-not (item-class nt)
X	    (finish-rhs
X	      (parse-while-lookahead-not
X		item-class nt grammar item-list)
X	      grammar (cdr rhs))]
X	  [separated-list (nt item-class)
X	    (finish-rhs
X	      (parse-separated-list
X		item-class nt grammar item-list)
X	      grammar (cdr rhs))]
X	  [else (error 'parse-rhs "bad rhs: ~s" rhs)]
X	  ))))
X
X(define finish-rhs
X  (lambda (result1 grammar rhs)
X    ;; run rhs against unused portion in result1, then cons together
X    ;; the resulting trees.
X    (let ([result2
X	    (parse-rhs grammar rhs
X	      (parser-result->unused result1))])
X      (make-parser-result
X	(cons (parser-result->tree result1)
X	      (parser-result->tree result2))
X	(parser-result->unused result2)))))
X
X(define parse-data-item
X  (lambda (item-class item-list)
X    ;; if (car item-list) belongs to item-class, create a
X    ;; parser-result containing its data field.  If not, report an
X    ;; error. 
X    (if (eq? (lexical-item->class (car item-list))
X	     item-class)
X      (make-parser-result
X	(lexical-item->data (car item-list))
X	(cdr item-list))
X      (error 'parse-rhs-data-item "Looking for: ~s found: ~s"
X	item-class
X	(lexical-item->class (car item-list))))))
X
X(define parse-while-lookahead-not
X  (lambda (item-class nt grammar item-list)
X    ;; returns parse tree showing list of parse trees.
X    (if
X      (eq? (lexical-item->class (car item-list))
X	   item-class)
X      ;; if the lookahead item matches the terminator, then we're
X      ;; done-- there are no elements in the resulting list.
X      (make-parser-result '() item-list)
X      ;; otherwise, run the rhs, recur on what's left over, and then
X      ;; cons the results together.
X      (let ([result1 (parse-nt grammar nt item-list)])
X	(let ([result2 (parse-while-lookahead-not
X			 item-class nt grammar
X			 (parser-result->unused result1))])
X	  (make-parser-result
X	    (cons (parser-result->tree result1)
X		  (parser-result->tree result2))
X	    (parser-result->unused result2)))))))
X
X(define parse-separated-list
X  (lambda (item-class nt grammar item-list)
X    (letrec
X      ([loop (lambda (item-list)
X	       ;; {item-class nt}*
X	       (if (not (eq? (lexical-item->class (car item-list))
X			     item-class))
X		 (make-parser-result
X		   '()
X		   item-list)
X		 (let ([result1 (parse-nt grammar nt (cdr item-list))])
X		   (let ([result2
X			   (loop (parser-result->unused result1))])
X		     (make-parser-result
X		       (cons
X			 (parser-result->tree result1)
X			 (parser-result->tree result2))
X		       (parser-result->unused result2))))))])
X      (let ([result1 (parse-nt grammar nt item-list)])
X	(let ([result2 (loop (parser-result->unused result1))])
X	  (make-parser-result
X	    (cons (parser-result->tree result1)
X		  (parser-result->tree result2))
X	    (parser-result->unused result2)))))))
X
X;;; ****************************************************
X
X;;; Sample grammar
X
X(define **keywords-list** '(begin2 begin end while do if then else))
X
X(define grammar-1
X  '((command
X      ((token begin2) (non-terminal command) 
X       (token semicolon) (non-terminal command)
X       (token semicolon) (token end))
X      compound-command)
X    (command
X      ((token begin)
X       (non-terminal command-list)
X       (token end))
X      sequence-command)
X    (command-list			; this makes semicolon a terminator
X      ()
X      *empty-list)
X    (command-list
X      ((non-terminal command)
X       (token semicolon)
X       (non-terminal command-list))
X      *command-list)
X    (command
X      ((token while) (non-terminal expression)
X       (token do) (non-terminal command))
X      while-command)
X    (command
X      ((token if) (non-terminal expression)
X       (token then) (non-terminal command)
X       (token else) (non-terminal command))
X      if-command)
X    (command
X      ((data-item identifier) (token assign-sym)
X       (non-terminal expression))
X      assignment-command)
X    (expression
X      ((data-item identifier)) identifier-expression)
X    (expression
X      ((token lparen) (non-terminal expression)
X       (token plus-sym) (non-terminal expression)
X       (token rparen))
X      sum-expression)))
X
X;;; Tests
X
X(define test1
X  (lambda () (parse "x := y")))
X
X(define test2
X  (lambda () (parse "begin2 x:= y ; y := zz; end")))
X
X(define test3
X  (lambda () (parse "begin x:=y; y:=z; z:=x; end")))
X
X; > (test1)
X; (assignment-command x (identifier-expression y))
X; > (test2)
X; (compound-command
X;    (assignment-command x (identifier-expression y))
X;    (assignment-command y (identifier-expression zz)))
X; > (test3)
X; (sequence-command
X;    (*command-list
X;       (assignment-command x (identifier-expression y))
X;       (*command-list
X;          (assignment-command y (identifier-expression z))
X;          (*command-list
X;             (assignment-command z (identifier-expression x))
X;             (*empty-list)))))
X
SHAR_EOF
if test 14401 -ne "`wc -c ll-parsers.s`"
then
echo shar: error transmitting ll-parsers.s '(should have been 14401 characters)'
fi
echo shar: extracting grammars.s '(7331 characters)'
sed 's/^X//' << \SHAR_EOF > grammars.s
X;;; process grammars
X
X;;; interface for translating sps grammars into form suitable for
X;;; LL parser-builder
X
X;;; top-level entry:  define-grammar
X
X(extend-syntax (define-grammar)
X  [(define-grammar grammar-name start-symbol
X     (prod-name (lhs rhs) vars action)
X     ...) 
X   (with ((ans
X	    (and*
X	      (check-grammar
X		'((prod-name (lhs rhs) vars action) ...)))))
X     (if 'ans
X       (let
X	 ((scanner (make-scanner '(rhs ...)))
X	  (parser (make-parser
X		   'start-symbol
X		   (list		; production in form for LL-builder
X		     (list 'lhs
X			   (translate-rhs 'rhs)
X			   (lambda vars action))
X		     ...))))
X	 (begin
X	   (set! scanner-2 scanner)
X	   (set! grammar-name
X	     (lambda (string)
X	       '(printf "input string: ~s~%" string)
X	       (let ((item-list (scanner string)))
X		 '(printf "item-list: ~s~%" item-list)
X		 (parser item-list))))
X	   (declare-unchecked grammar-name () (-> (seq string) start-symbol))
X	   (printf "grammar ~s checked~%" 'grammar-name)))))])
X
X(extend-syntax (define-unchecked-grammar)
X  [(define-unchecked-grammar grammar-name start-symbol
X     (prod-name (lhs rhs) vars action)
X     ...)
X   (let
X	 ((scanner (make-scanner '(rhs ...)))
X	  (parser (make-parser
X		   'start-symbol
X		   (list		; production in form for LL-builder
X		     (list 'lhs
X			   (translate-rhs 'rhs)
X			   (lambda vars action))
X		     ...))))
X	 (begin
X	   (set! scanner-2 scanner)
X	   (set! grammar-name
X	     (lambda (string)
X	       '(printf "input string: ~s~%" string)
X	       (let ((item-list (scanner string)))
X		 '(printf "item-list: ~s~%" item-list)
X		 (parser item-list))))
X	   'grammar-name))])
X
X(define make-scanner
X  (lambda (rhss)
X    (let*
X      ((keywords (collect-keywords rhss))
X       (cook-symbol
X	 (lambda (chars)
X	   (let ([sym (char-list->symbol chars)])
X	     (if (memq sym keywords)
X	       (make-lexical-item sym '())
X	       (make-lexical-item 'symbol sym)))))
X       (automaton
X	 `((initial ,lex-error
X	     ((#\space whitespace)
X	      (#\tab  whitespace)
X	      (#\newline whitespace)
X	      ((#\a #\z) symbol)
X	      ((#\A #\Z) symbol)
X	      (#\, comma)
X	      ((#\* #\/) symbol)	; except "," 
X	      (#\= symbol)
X	      (#\( lparen)
X	      (#\) rparen)
X	      (#\; semicolon)
X	      ((#\0 #\9) number)))
X	   (symbol ,cook-symbol
X	     (((#\a #\z) symbol)
X	      ((#\A #\Z) symbol)
X	      ((#\0 #\9) symbol)
X	      (#\* symbol)
X	      (#\+ symbol)
X	      (#\- symbol)
X	      (#\/ symbol)
X	      (#\! symbol)
X	      (#\= symbol)))
X	   (whitespace ,cook-bogus
X	     ((#\space whitespace)
X	      (#\tab whitespace)
X	      (#\newline whitespace)))
X	   (number ,cook-number
X	     (((#\0 #\9) number)))
X	   (comma ,(cook-token 'comma) ())
X	   (lparen ,(cook-token 'lparen) ())
X	   (rparen ,(cook-token 'rparen) ())
X	   (semicolon ,(cook-token 'semicolon) ())
X	   (end-marker ,(cook-token 'end-marker) ()))))
X      (set! automaton-2 automaton)
X      (lambda (string) (scan-string automaton string)))))
X
X(define collect-keywords
X  (lambda (rhss)
X    (if (null? rhss) '()
X      (let loop
X	((rhs (car rhss))
X	 (keywords (collect-keywords (cdr rhss))))
X	'(printf "collect-keywords-loop: rhs = ~s keywords = ~s rhss = ~s~%"
X	  rhs keywords rhss)
X	(cond
X	  ((null? rhs) keywords)
X	  ((symbol? (car rhs))		; it's a non-terminal
X	   (loop (cdr rhs) keywords))
X	  ((eq? (caar rhs) 'quote)	; it's a terminal symbol
X	   '(printf "looking at terminal symbol ~s~%" (car rhs))
X	   (loop (cdr rhs)
X		 (if
X		   (or (memq (cadar rhs) '(lparen rparen comma semicolon))
X		       (memq (cadar rhs) keywords))
X		   keywords
X		   (cons (cadar rhs) keywords))))
X	  (else (report-error
X		  (list "unknown grammar item" (car rhs)))))))))
X
X;;; ****************************************************************
X
X;;; Translation to LL parsing system
X
X;;; production ::= (prod-name (lhs rhs) action)  
X;;; action ::= a closure (of several arguments)
X
X(define make-parser
X  (lambda (start-symbol productions)
X    (let ((table (build-LL-table start-symbol productions)))
X      (set! table-2 table)
X      (lambda (item-list)
X	(let ([result (parse-nt table start-symbol item-list)])
X	  (if (eq? (lexical-item->class
X		     (car (parser-result->unused result)))
X		   'end-marker)
X	    (parser-result->tree result)
X	    (error 'parse-top-level
X	      "symbols left over: ~s"
X	      (car (parser-result->unused result)))))))))
X
X(define translate-rhs
X  (lambda (rhs)
X    (map translate-rhs-item rhs)))
X
X(define translate-rhs-item
X  (lambda (item)
X    (cond
X      ((memq item '(int symbol))	; it's one of the data-items
X       `(data-item ,item))
X      ((symbol? item) `(non-terminal ,item))
X      ((eq? (car item) 'quote)		; it's a token
X       `(token ,(cadr item)))
X      (else (report-error
X	      (list "unknown grammar item" item))))))
X
X;;; ****************************************************************
X
X;;; Tests
X
X'(define-unchecked-grammar sdt2 pgm
X  (pgm
X    (pgm (exp))
X    (exp)
X    `(pgm ,exp))
X  (static
X    (exp (symbol))
X    (lit)
X    `(ident ,lit))
X  (if
X    (exp ('if exp exp exp))
X    (exp0 exp1 exp2)
X    `(cond ,exp0 ,exp1 ,exp2))
X  (appl
X    (exp ('lparen exp exp 'rparen))
X    (e0 e1)
X    `(appl ,e0 ,e1)))
X
X;;; ****************************************************************
X
X;;; Interface to typechecker
X
X(define check-grammar
X  (lambda (productions)
X    (map
X      (lambda (production)
X	(call/cc (lambda (next)
X		   (check-production production
X		     (lambda (msg)
X		       (printf "~s:~%" (car production))
X		       (pretty-print msg)
X		       (next #f)))
X		   (printf "~s~%" (car production))
X		   #t)))
X      productions)))
X
X(define check-production
X   (lambda (production fail)
X      (let
X	 ((prod-name (gen-fcn-sym))
X	  (type-info (filter-type-info (cadr (cadr production))))
X					; get type info for rhs
X	  (vars (caddr production))
X	  (term (cadddr production)))
X	; (printf "check-production: production = ~s~%" production)
X	; (printf "prod-name = ~s type-info = ~s vars = ~s term = ~s~%"
X        ;	   prod-name type-info vars term)
X	 '(add-undefined-types-to-badlist (cons 'dummy type-info) nil)
X	 (checker
X	    (ext-prefix vars type-info
X	       (ext-prefix (list prod-name)
X		 (list
X		   (make-type-scheme
X		   '()
X		   (make-functional-type
X		     (list (car (cadr production))) ; lhs nonterminal
X						    ; is also a type.
X		     (gen-type-var))))
X		 empty-prefix))
X	    (make-application prod-name (list term))
X	    empty-prefix
X	    fail))))
X
X(define filter-type-info
X  (lambda (l)
X    (cond
X      ((null? l) nil)
X      ((and
X	 (not (atom? (car l)))
X	 (eq? (caar l) 'quote))
X       (filter-type-info (cdr l)))
X      (t (cons
X	   (make-type-scheme
X	     '()
X	     (car l))
X	   (filter-type-info (cdr l)))))))
X
X(define check-production1
X  (lambda (production)
X    (let ((ans (check-production production report-error)))
X      (pretty-print (expand-type-exp (car ans) (cadr ans))))))
X
X(define and*
X  (lambda (l)
X    (cond
X      ((null? l) #t)
X      ((car l) (and* (cdr l)))
X      (else #f))))
X
X;;; ****************************************************************
X
X;;; A test tag, currently commented out:
X
X'(define-type-abbrev Num int)
X'(define-type-abbrev Digit int)
X
X'(define-grammar sdt3 Num
X  (zero
X    (Digit ('a))
X    ()
X    0)
X  (one
X    (Digit ('b))
X    ()
X    1)
X  (empty-number
X    (Num ())
X    ()
X    0)
X  (non-empty-number
X    (Num (Digit Num))
X    (x y)
X    (+ (* 2 y) x)))
X    
SHAR_EOF
if test 7331 -ne "`wc -c grammars.s`"
then
echo shar: error transmitting grammars.s '(should have been 7331 characters)'
fi
echo shar: extracting Makefile '(693 characters)'
sed 's/^X//' << \SHAR_EOF > Makefile
Xall1 = top.s chez-support.s data-structures.s mgu.s check-exp.s
Xall2 = defining-forms.s type-constructors.s globals.s
Xall3 = scanner.s ll-parsers.s grammars.s Makefile README NEWS
Xall = $(all1) $(all2) $(all3)
X
Xexamples = sem2.s decls-example.s
X
Ximprint:
X	imprint -Pimagen -2 -O $(all)
X
Xdoc:
X	tex doc.tex
X	iptex doc.dvi
X
Xwc:
X	wc $(all)
X
Xarchive:
X	ar ruv arch $(all) *.tex $(examples)
X	split arch arch.
X
Xshar:
X	shar -pX -cv $(all1) > shar.1
X	shar -pX -cv $(all2) > shar.2
X	shar -pX -cv $(all3) > shar.3
X	shar -pX -cv *.tex > shar.4
X	shar -pX -cv $(examples) > shar.5
X
Xextract:
X	mv data-structures data-structures.s
X	mv defining-forms. defining-forms.s
X	mv type-constructo type-constructors.s
X
X
SHAR_EOF
if test 693 -ne "`wc -c Makefile`"
then
echo shar: error transmitting Makefile '(should have been 693 characters)'
fi
echo shar: extracting README '(1310 characters)'
sed 's/^X//' << \SHAR_EOF > README
X-*-text-*-
X
XSPS
X
XSPS Version 1.4 (Chez Scheme Version) Mon Apr 17 22:01:21 1989
X
XThis is the Semantic Prototyping System (SPS).  SPS is a suite of
Xprograms, written in Scheme, for rapid prototyping and testing of
Xlanguage definitions written using denotational semantics.  It
Xconsists of an ML-style typechecker for a large subset of Scheme,
Xalong with an LL(1) parser and transducer generator.
X
XThe system is shipped in several files:
X
XREADME		this file
Xshar.1 -- shar.5	shar archive files
X
XTo build the system, remove the mail headers from each of the archive files,
Xand then do 
X
Xsh shar.1
X..
Xsh shar.5
X
XThis should build all the right files.  See the Makefile for a list of files
Xthat should be present.
X
XThe current version of SPS is written in Chez Scheme.  Load it as follows:
X
X  1.  Load Chez scheme.	
X  3.  Execute (load ".....top.s")    where .... is the path name for the
Xdirectory where SPS resides.
X
XIf you intend to use the tools system seriously, it is recommended that you
Xre-boot scheme, loading the types system before dumping.  The resulting heap
Ximage can be loaded by using Chez Scheme with the -h option in the command
Xline.  
X
XMitchell Wand
XCollege of Computer Science
XNortheastern University
X360 Huntington Avenue #161CN
XBoston, MA 02115
X
XCSNet:  wand@corwin.ccs.northeastern.edu
X
X
X
X
X
X
SHAR_EOF
if test 1310 -ne "`wc -c README`"
then
echo shar: error transmitting README '(should have been 1310 characters)'
fi
echo shar: extracting NEWS '(1666 characters)'
sed 's/^X//' << \SHAR_EOF > NEWS
X-*-text-*-
X
XSPS NEWS:  Changes since previous versions
X
XChanges in SPS version 1.6 (April, 1989)
X
X1.  type literal is now type symbol.  symbols are created by expressions of
Xthe form (quote <symbol>), and by reading the non-terminal symbol in the
Xparser generator.
X
X2. (heterogeneous) is now called (unchecked).  Values of type unchecked are
Xproduced by the the special form (unchecked ....), which takes an arbitrary
Xargument (not processed by the checker).
X
X3. The Scheme function (list ...) now works the way it should.
X
X4. casefn is now called union-case.  A comment has been added to the manual
Xsuggesting that most of the time one should use define-record-type instead.
X
X5. Code for detecting illegal cycles of type-abbrevs has been removed.  These
Xare no longer detected.
X
X6. define-record-type now calculates its own name for the case function,
Xinstead of taking a name as an argument.
X
X7. The syntax for define-type-constructor has changed.  The pattern is now:
X
X(define-type-constructor
X     (name type-var ...)
X     rep-type
X     (fn-name fn-type value) ...)
X
XThis syntax is changed slightly from old versions:  it declares the
Xglobal generics the same way as define-record-type, and individual
Xfunctions must have all their generics declared locally with
X(generic (x y) ...) .
X
X8. This version includes an LL(1) parser generator.  So some grammars which
Xpreviously worked will have to be changed.  Consequently, the top-level
Xentries for grammars have changed entirely.  See define-grammar in doc.txt .
XAlso, the format for concrete productions has changed:  the right-hand side is
Xnow enclosed in an extra set of parens.  See doc.txt for an example.
X
X
X
SHAR_EOF
if test 1666 -ne "`wc -c NEWS`"
then
echo shar: error transmitting NEWS '(should have been 1666 characters)'
fi
#	End of shell archive
exit 0
